// Facet.cc

#include <iostream>
#include <cstdlib>
#include <NTL/mat_ZZ.h>

#include "Voronoi.h"
#include "Facet.h"
#include "la_support.h"

void die_facet(const char* last_words)
{
  std::cout <<"Facet.cc: " <<last_words;
  exit(EXIT_FAILURE);
}

Facet::Facet(double** in)
{
  doubleMakeSquareMatrix(dim+1, &in_simplex);
  doubleMakeVector(dim+1, &out_vertex);
  doubleMakeVector(dim+1, &alpha);
  doubleCopySquareMatrix(dim+1, in, in_simplex);
}

Facet::Facet(const Facet& F)
{
  doubleMakeSquareMatrix(dim+1, &in_simplex);
  doubleMakeVector(dim+1, &out_vertex);
  doubleMakeVector(dim+1, &alpha);
  doubleCopySquareMatrix(dim+1, F.in_simplex, in_simplex);
  doubleCopyVector(dim+1, F.out_vertex, out_vertex);
  doubleCopyVector(dim+1, F.alpha, alpha);
  det = F.det;
}

Facet::~Facet()
{
  doubleDestroySquareMatrix(dim+1, &in_simplex);
  doubleDestroyVector(dim+1, &out_vertex);
  doubleDestroyVector(dim+1, &alpha);
}

Facet& Facet::operator=(const Facet& F)
{
  if (this != &F) {
    doubleMakeSquareMatrix(dim+1, &in_simplex);
    doubleMakeVector(dim+1, &out_vertex);
    doubleMakeVector(dim+1, &alpha);
    doubleCopySquareMatrix(dim+1, F.in_simplex, in_simplex);
    doubleCopyVector(dim+1, F.out_vertex, out_vertex);
    doubleCopyVector(dim+1, F.alpha, alpha);
    det = F.det;
  }
  return *this;
}

bool Facet::operator<(const Facet& F) const
{
  for (int i = 1; i < dim; i++)
    for (int j = 0; j < dim; j++) {
      double sign = in_simplex[i][j]-F.in_simplex[i][j];
      if (sign < -0.5) return true;
      if (sign > 0.5) return false;
    }
  return false;
}

bool Facet::operator==(const Facet& F) const
{
  return !(*this < F) && !(F < *this);
}

void Facet::insert_out_vertex(double* out)
{
  // consistency check
  for (int i = 0; i < dim+1; i++)
    if (out_vertex[i] != 0.0)
      die_facet("out_vertex[i] != 0.0.\n");

  doubleCopyVector(dim+1, out, out_vertex);

  NTL::mat_ZZ mat;
  mat.SetDims(dim+1,dim+1);
  for (int i = 0; i < dim+1; i++)
    for (int j = 0; j < dim+1; j++)
      mat[i][j] = NTL::to_ZZ(in_simplex[i][j]);
  
  NTL::vec_ZZ rhs;
  rhs.SetLength(dim+1); 
  for (int i = 0; i < dim+1; i++)
    rhs[i] = NTL::to_ZZ(out_vertex[i]);

  NTL::ZZ det_zz;
  NTL::vec_ZZ alpha_zz;

  solve(det_zz, alpha_zz, mat, rhs, true);
  
  for (int i = 0; i < dim+1; i++) 
    if (det_zz > 0) 
      alpha[i] = NTL::to_double(alpha_zz[i]);
    else
      alpha[i] = -NTL::to_double(alpha_zz[i]);

  det = fabs(NTL::to_double(det_zz));
}
