// Regulator.cc

#include <cmath>
#include <iostream>

#include <cstdlib>

#include "Regulator.h"
#include "Voronoi.h"
#include "la_support.h"

void die_regulator(const char* last_words)
{
  std::cout <<"Regulator.cc: " <<last_words;
  exit(EXIT_FAILURE);
}

int gcd(int a, int b)
{
  if (a < b)
    return gcd(b,a);
  else {
    int r = a % b;
    if (r == 0) 
      return b;
    else 
      return gcd(b,r);
  }
}

Regulator::Regulator(const Facet& F)
{
  doubleMakeVector(Dim, &q);

  for (int k = 0; k < dim; k++) 
    for (int l = 0; l < k+1; l++)
      if (k == l) {
	q[k*(k+1)/2+l] = F.det*F.out_vertex[k]*F.out_vertex[l]; 
	for (int i = 0; i < dim+1; i++)
	  q[k*(k+1)/2+l] = q[k*(k+1)/2+l] - F.alpha[i]*F.in_simplex[i][k]*F.in_simplex[i][l];
      }
      else {
	q[k*(k+1)/2+l] = 2.0*F.det*F.out_vertex[k]*F.out_vertex[l];
	for (int i = 0; i < dim+1; i++)
	  q[k*(k+1)/2+l] = q[k*(k+1)/2+l] - 2.0*F.alpha[i]*F.in_simplex[i][k]*F.in_simplex[i][l];
      }
}

Regulator::Regulator(const Regulator& R)
  : facets(R.facets)
{
  doubleMakeVector(Dim, &q);
  doubleCopyVector(Dim, R.q, q);
}

Regulator::~Regulator()
{
  doubleDestroyVector(Dim, &q);
  facets.clear();
}

Regulator& Regulator::operator=(const Regulator& R)
{
  if (this != &R) {
    doubleCopyVector(Dim, R.q, q);
    facets = R.facets;
  }
  return *this;
}

bool Regulator::operator==(const Regulator& R) const
{
  return !(*this < R) && !(R < *this);
}

bool Regulator::operator<(const Regulator& R) const
{
  for (int i = 0; i < Dim; i++) {
    double sign = q[i] - R.q[i];
    if (sign < -0.5) return true;
    if (sign > 0.5) return false;
  }
  return false;
}

void Regulator::insert_facet(const Facet* F)
{
  facets.push_front(const_cast<Facet*>(F));
}

