// scc.cc
// version 2.0

#include <deque>
#include <vector>
#include <list>
#include <iostream>
#include <unistd.h>
#include <NTL/ZZ.h>

#include "Voronoi.h"
#include "DeloneTriangulation.h"
#include "la_support.h"

#include <cstdio>

#include <time.h>

// Get current time, format is HH:mm:ss
// Slightly modified http://stackoverflow.com/questions/997946/how-to-get-current-time-and-date-in-c
const std::string currentTime() {
    time_t     now = time(0);
    struct tm  tstruct;
    char       buf[80];
    tstruct = *localtime(&now);
    strftime(buf, sizeof(buf), "%X", &tstruct);

    return buf;
}

int next_subset(int quan, int total, int *a){
/* Finding the next subset of quan elements. Starting subset is quan ones in the beginning */
int i,j,k;

//int result;

i=0;

while((a[i]==1)&&(i<total)){
  i++;
}
j=i;

while((a[i]==0)&&(i<total)){
  i++;
}
//printf("quan=%d, total=%d, j=%d, i=%d\n",quan,total,j,i);

if((i==total)&&(j==0)){
  for(k=0;k<quan;k++)
	a[k]=1;
  return 1;
}

if(j==total)
  return 0;

if((j==0)&&(i==(total-quan)))
  return 0;

if(j!=0){
  a[j]=1;
  a[j-1]=0;
  return 1;
}

while((a[j+i]==1)&&(i+j<total)){
  j++;
}
a[i+j]=1;

for(k=0;k<j-1;k++)
  a[k]=1;

for(k=j-1;k<i+j;k++)
  a[k]=0;

return 1;
}

QForm make_form(int marker_size){
	QForm temp;
//	temp.codim=-1;
	doubleMakeVector(Dim,&temp.q);
	intMakeVector(dim,&temp.ranks);
	intMakeVector(marker_size,&temp.marker);
	return temp;
}

/*void write_form_file(QForm form, FILE* form_file){
	fprintf(form_file,"%d\n\n",dim);
	for(int i=0;i<dim;i++){
		for(int j=0; j<=i;j++){
			fprintf(form_file,"%4d",(int)form.q[i*(i+1)/2+j]);
		}
		fprintf(form_file,"\n");
	}
}*/

int dim;
int Dim;
int optimal_hash_table_size;

void die_scc(char* last_words)
{
  std::cout <<"scc.cc: " <<last_words;
  exit(EXIT_FAILURE);
}

//

int main(int argc, char** argv)
{

  bool random_walk_option = false;
  bool greedy_walk_option = false;
  bool read_input_option = false;
  bool write_coop_option = false;
  bool full_classification = false;
  bool write_form_option = false;

  int number_random_steps = 0;

  int c;
  dim = 0;

  while ((c = getopt (argc, argv, "ahgivwd:r:")) != -1)
    switch (c)
      {
      case 'h':
	std::cout <<"Usage: scc [options] <file\n";
	std::cout <<"Options:\n";
	std::cout <<" -h  show this help\n";
	std::cout <<" -dD dimension D (required)\n";
	std::cout <<" -a find all Delone decompositions\n";
	std::cout <<" -rN cruise randomly (N times)\n";
	std::cout <<" -g  cruise greedily\n";
	std::cout <<" -i  read coop file\n";
	std::cout <<" -w  write coop file for every secondary cone\n";
	std::cout <<" -v  write central form information for every cone\n";
	return 0;
	break;
      case 'd':
	dim = atoi(optarg);
	Dim = dim*(dim+1)/2;
	optimal_hash_table_size = factorial(dim);
	break;
      case 'g':
	greedy_walk_option = true;
	break;
      case 'r':
	random_walk_option = true;
	number_random_steps = atoi(optarg);
	break;
      case 'i':
	read_input_option = true;
	break;
      case 'w':
	write_coop_option = true;
	break;
      case 'a':
	full_classification = true;
	break;
      case 'v':
	write_form_option = true;
	break;
      default:
	die_scc("invalid option\nTry 'scc -h' for more information.\n");
	return 1;
      }

  if (dim <= 0)
    die_scc("invalid options\noption '-d' is needed.\n");

  std::cout <<"*** sc ***\n";
  std::cout <<"***  c ***\n\n";

  // prepare starting triangulation

  DeloneTriangulation Del;
  if (read_input_option) {
    Del.read_COOP_file(stdin);
  }
  else
    Del = principal_domain();
  
  std::vector<DeloneTriangulation> classification;

  classification.reserve(10000000);

  std::list<int> known_adj; 
  std::deque< std::pair<int, std::list<int> > > todo;

  classification.push_back(Del);
  todo.push_back(make_pair(0, known_adj));

  //
  //
  //

  if (random_walk_option) {

    for (int current = 0; current < number_random_steps; current++) {
      NTL::ZZ invariant;
      double lower_bound;

      Del.compute_lower_bound(invariant, lower_bound);

//      Del.compute_invariant(invariant);

      std::cout << "Time: " << currentTime() << std::endl;

      std::cout <<"Looking at #" <<current <<" (normalized covering density >= " <<lower_bound <<")\n";

//      std::cout <<"Looking at #" <<current <<"\n";

      int no_neighbours = Del.compute_no_neighbours();
      std::cout <<" * There are " <<no_neighbours <<" neighbours.\n"; 

      if (write_coop_option) {
	std::cout <<" * Write output for COOP " <<std::endl;    
	FILE* coop_file;
	char file_name[100];
	sprintf(file_name, "sc%d_%d.coop", dim, current);
	coop_file = fopen(file_name, "w");
	Del.write_COOP_file(coop_file);
	fclose(coop_file);
      }


/*      int simpl = Del.no_simplices(); // A.G.
	  std::cout << " * Number of simplices classes: " << simpl << std::endl; // A.G.

      int vol = Del.compute_max_volume(); // A.G.
	  std::cout << " * Maximal simplex volume: " << vol << std::endl; // A.G.*/

      int r = rand() % no_neighbours;
      std::cout <<" * Proceed with number " <<r <<".\n\n";
      DeloneTriangulation random_neighbour = Del.compute_nth_neighbour(r);
      Del = random_neighbour;
    }
  }


  else if (greedy_walk_option) {
    int current = 0;
    bool finished = false;

    while (!finished) {
      NTL::ZZ invariant;
      double lower_bound;

      Del.compute_lower_bound(invariant, lower_bound);

//      Del.compute_invariant(invariant);

      std::cout << "Time: " << currentTime() << std::endl;

      std::cout <<"Looking at #" <<current <<" (normalized covering density >= " <<lower_bound <<")\n";

//      std::cout <<"Looking at #" <<current <<"\n";

      int no_neighbours = Del.compute_no_neighbours();
      std::cout <<" * There are " <<no_neighbours <<" neighbours.\n"; 

      if (write_coop_option) {
	std::cout <<" * Write output for COOP " <<std::endl;    
	FILE* coop_file;
	char file_name[100];
	sprintf(file_name, "sc%d_%d.coop", dim, current);
	coop_file = fopen(file_name, "w");
	Del.write_COOP_file(coop_file);
	fclose(coop_file);
      }

      std::vector<DeloneTriangulation> neighbours(no_neighbours);
      for (int i = 0; i < no_neighbours; i++)
	neighbours[i] = Del.compute_nth_neighbour(i);
      
      int number_best = -1;
      Del.compute_lower_bound(invariant, lower_bound);
      double least_lower_bound = lower_bound;
      for (int i = 0; i < no_neighbours; i++) {
	neighbours[i].compute_lower_bound(invariant, lower_bound);
	if (lower_bound < least_lower_bound) {
	  least_lower_bound = lower_bound;
	  number_best = i;
	}
      }
      
      finished = (number_best == -1);
      if (!finished) {
	current++;
	Del = neighbours[number_best];
	std::cout <<"\n";
      }
    }
  }      

  else {

    while (!todo.empty()) {
      int current = todo.front().first; // current number in the classification
      Del = classification[current]; 	// current triangulation
      known_adj = todo.front().second;	// known adjacent triangulations
      todo.pop_front();					// pop current triangulation from todo double queue

      NTL::ZZ invariant;
      double lower_bound;

//      Del.compute_lower_bound(invariant, lower_bound);

      Del.compute_invariant(invariant);

	  std::cout << "Time: " << currentTime() << std::endl;
    
//      std::cout <<"Looking at #" <<current <<" (normalized covering density >= " <<lower_bound <<")\n";

      std::cout <<"Looking at #" <<current <<"\n";

/*      int simpl = Del.no_simplices(); // A.G.
	  std::cout << " * Number of simplices classes: " << simpl << std::endl; // A.G.

      int vol = Del.compute_max_volume(); // A.G.
	  std::cout << " * Maximal simplex volume: " << vol << std::endl; // A.G.*/


      std::cout <<" * Remaining: " <<todo.size() <<std::endl;    



      int no_neighbours = Del.compute_no_neighbours();

      if (write_coop_option) {
	std::cout <<" * Write output for COOP " <<std::endl;    
	FILE* coop_file;
	char file_name[100];
	sprintf(file_name, "sc%d_%d.coop", dim, current);
	coop_file = fopen(file_name, "w");
	Del.write_COOP_file(coop_file);
	fclose(coop_file);
      }

      std::vector<DeloneTriangulation> neighbours(no_neighbours);
      for (int i = 0; i < no_neighbours; i++)
	neighbours[i] = Del.compute_nth_neighbour(i);

      std::cout <<" * There are " <<no_neighbours <<" neighbours.\n"; 
      std::cout <<" * ";

      int no_extreme = Del.compute_no_extreme();

      for (int i = 0; i < no_neighbours; i++) {

	std::cout <<"(" <<i <<std::flush;
	
	// 1. check if neighbour[i] = Del
	bool known = neighbours[i].is_equivalent(Del);
	if (known) 
	  std::cout <<"=" <<current <<") " <<std::flush;
	
	// 2. check if neighbour[i] = completed where adjacencies are known
	if (!known) {
	  // use gathered information for adjacency-tests to completed
	  std::list<int>::iterator pos = known_adj.begin(); 
	  while (pos != known_adj.end() && !known) {
	    known = neighbours[i].is_equivalent(classification[*pos]);
	    if (known) 
	      std::cout <<"=" <<*pos <<") " <<std::flush;
	    pos++;
	  }
	}

	// 3. check if neighbour[i] = todo[j]
	if (!known) {
	  for (unsigned int j = 0; j < todo.size() && !known; j++) {
	    known = neighbours[i].is_equivalent(classification[todo[j].first]);
	    if (known) {
	      // here we know that todo[j] is adjacent to Del (current)
	      if (find(todo[j].second.begin(), todo[j].second.end(), current) == todo[j].second.end()) 
		todo[j].second.push_back(current);
	      std::cout <<"=" <<todo[j].first <<") " <<std::flush;
	    }
	  }
	}

	// 4. All checks failed. Hence we have a new type.
	if (!known) {
	  // here we know at least that neighbours[i] is adjacent to Del.
	  std::list <int> new_adj;
	  new_adj.push_back(current);
	  classification.push_back(neighbours[i]);
	  todo.push_back(make_pair(classification.size()-1, new_adj));
	  std::cout <<"->" <<classification.size()-1 <<") " <<std::flush;
	}
      }
      std::cout <<"\n\n";
    }

    std::cout <<"--\n";
    std::cout <<"Classification of Delone triangulations completed!\n";
    std::cout <<"There are " <<classification.size() 
	      <<" non-equivalent Delone triangulations in dimension " <<dim <<".\n";

	if(full_classification)
	{
	  std::vector<QForm> full_cl;
	  std::vector<int> source_tr; // Number of the triangulation where this tiling appears for the first time
	  QForm temp_form;
	  temp_form.q=(double*)malloc(Dim*sizeof(double));
	  temp_form.ranks=(int*)malloc(dim*sizeof(int));
	  std::cout <<"\n";
	  std::cout <<"Starting classification of arbitrary Delone decompositions\n";
//	  std::cout << "Time: " << currentTime() << std::endl;
	  int i;
	  int marker[1000];

	  full_cl.reserve(10000000);


	  for(i=0;i<classification.size();i++){
		classification[i].compute_no_neighbours();
		int max_extreme=classification[i].compute_no_extreme();
		for(int ii=0;ii<max_extreme;ii++)
		  marker[ii]=1;
		classification[i].subset(marker,&temp_form);

	    QForm push_form;
		push_form=make_form(max_extreme);
//				  push_form.codim=temp_form.codim;
		for(int ii=0;ii<dim;ii++)
		  push_form.ranks[ii]=temp_form.ranks[ii];
		for(int ii=0;ii<Dim;ii++)
		  push_form.q[ii]=temp_form.q[ii];
		for(int ii=0;ii<max_extreme;ii++)
		  push_form.marker[ii]=marker[ii];
		push_form.parent=i;
		push_form.codim=0;
		full_cl.push_back(push_form);
		source_tr.push_back(i);
	  }
	  int imin=0;
	  int imax=full_cl.size();

	  if(write_form_option){
		for(int ii=imin;ii<imax;ii++){
		  FILE* form_file;
		  char file_name[100];
		  sprintf(file_name, "sc%d_%d.form", dim, ii);
		  form_file = fopen(file_name, "w");
		  classification[full_cl[ii].parent].write_form_file(full_cl[ii],form_file);
		  fclose(form_file);
		}	
	  }

	  std::cout <<"There are " <<imax-imin<<" forms of codimension 0\n";
	  std::cout << "Time: " << currentTime() << std::endl;

	  for(i=1;(i<Dim)&&(imin<imax);i++){
		for(int j=imin;j<imax;j++){
		  int* cand_marker;
		  int cand_num;
		  cand_num=(classification[full_cl[j].parent]).subset_facets(i,full_cl[j].marker,&cand_marker);

//		  printf("codimension=%d cand_num=%d\n",i,cand_num);
//		  printf("cand_marker:");
		  
		  int max_extreme=classification[full_cl[j].parent].compute_no_extreme();

//		  for(int ii=0; ii< max_extreme*cand_num;ii++)
//			printf(" %d",cand_marker[ii]);
//		  printf("\n");

		  for(int k=0;k<cand_num;k++){
			for(int ii=0;ii<max_extreme;ii++){
			  marker[ii]=cand_marker[k*max_extreme+ii];			  
			}
			(classification[full_cl[j].parent]).subset(marker,&temp_form);
			bool found=false;

// Rank test

			double** mat;
			doubleMakeSquareMatrix(dim,&mat);
			for(int ii=0;ii<dim;ii++)
			  for(int jj=0;jj<dim;jj++){
			    if(ii<jj){
			      mat[ii][jj]=temp_form.q[jj*(jj+1)/2+ii];
			    }
			    else{
				  mat[ii][jj]=temp_form.q[ii*(ii+1)/2+jj];
			    }
 			  }
		
		    if(doubleMatrixRank(dim,mat)!=dim){
			  found=true;
			}
			doubleDestroySquareMatrix(dim,&mat);

// ISOM test with all previous of the same codimension

			int jj=imax;

		    while((!found)&&(jj<full_cl.size())){
			  bool ranks_ok = true;

			  for(int l=0; l<dim-1; l++)
				if(temp_form.ranks[l]!=full_cl[jj].ranks[l])
				  ranks_ok = false;
			  if(ranks_ok)
				found=isom(temp_form.q,full_cl[jj].q);
			  jj++;
		    }

		    if(!found){
			  QForm push_form;
			  push_form=make_form(max_extreme);
			  for(int ii=0;ii<dim;ii++)
			    push_form.ranks[ii]=temp_form.ranks[ii];
			  for(int ii=0;ii<Dim;ii++)
			    push_form.q[ii]=temp_form.q[ii];
			  for(int ii=0;ii<max_extreme;ii++)
				push_form.marker[ii]=marker[ii];
			  push_form.parent=full_cl[j].parent;
			  push_form.codim=i;
			  full_cl.push_back(push_form);
		    }
		  }
		  if(cand_num!=0)
			free(cand_marker);
		}
		imin=imax;
		imax=full_cl.size();

	  if(write_form_option){
		for(int ii=imin;ii<imax;ii++){
		  FILE* form_file;
		  char file_name[100];
		  sprintf(file_name, "sc%d_%d.form", dim, ii);
		  form_file = fopen(file_name, "w");
		  classification[full_cl[ii].parent].write_form_file(full_cl[ii],form_file);
		  fclose(form_file);
		}	
	  }


		std::cout <<"There are " << imax-imin <<" forms of codimension "<< i <<"\n";
		std::cout << "Time: " << currentTime() << std::endl;
	  }


	  std::cout <<"\nNumber of differrent decompositions: " <<full_cl.size()<<"\n";

	  std::cout <<"*******\n\n";
	}
  }
  
  return 0;
}
